"use strict";
var escapeRegExp = require('lodash.escaperegexp');
var clone = require('clone');
var compareObject = require('./compareObject');
var normalizeUrl = require('./normalizeUrl');
var PARSE_STATE = Object.freeze({
  PLAIN: 0,
  PLACEHOLDER: 1
});
var VERIFY = Object.freeze({NO_SLASH: 1});
var matcherPlaceholderDefault = '([^\/\?]+?)';
var Placeholder = function() {
  function Placeholder() {
    this.objectPath = null;
    this.urlMatcherGroupIndex = -1;
    this.verifyMode = 0;
    this.index = 0;
    this.belongsToCustomObject = null;
  }
  return ($traceurRuntime.createClass)(Placeholder, {verifyValue: function(value) {
      if (typeof value !== 'string' && typeof value !== 'number') {
        return false;
      }
      switch (this.verifyMode) {
        case VERIFY.NO_SLASH:
          if (typeof value === 'string' && value.indexOf('/') >= 0) {
            return false;
          }
      }
      return true;
    }}, {});
}();
var CustomObject = function() {
  function CustomObject(obj) {
    this.create = obj.create;
    this.isValid = obj.isValid;
    this.serializeValue = obj.serializeValue;
    this.objectPath = null;
    this.placeholders = [];
    this.index = null;
  }
  return ($traceurRuntime.createClass)(CustomObject, {addPlaceholder: function(placeholder) {
      this.placeholders.push(placeholder);
    }}, {});
}();
function initializeCustomObjects(template, customObjects, keyStack) {
  var keys = template ? Object.keys(template) : [];
  ++keyStack.length;
  var $__7 = true;
  var $__8 = false;
  var $__9 = undefined;
  try {
    for (var $__5 = void 0,
        $__4 = (keys)[$traceurRuntime.toProperty(Symbol.iterator)](); !($__7 = ($__5 = $__4.next()).done); $__7 = true) {
      var key = $__5.value;
      {
        var value = template[key];
        keyStack[keyStack.length - 1] = key;
        if (value && typeof value.create === 'function' && typeof value.isValid === 'function' && typeof value.serializeValue === 'function') {
          var customObject = new CustomObject(value);
          customObject.objectPath = keyStack.slice();
          customObject.index = customObjects.length;
          customObjects.push(customObject);
          template[key] = customObject;
        }
        if (typeof value === 'object') {
          initializeCustomObjects(value, customObjects, keyStack);
        }
      }
    }
  } catch ($__10) {
    $__8 = true;
    $__9 = $__10;
  } finally {
    try {
      if (!$__7 && $__4.return != null) {
        $__4.return();
      }
    } finally {
      if ($__8) {
        throw $__9;
      }
    }
  }
  --keyStack.length;
}
function verifyCompareObjectValue(templateValue, value) {
  if (templateValue instanceof CustomObject) {
    return !!templateValue.isValid(value);
  }
  return null;
}
function indexOfPlaceholder(value) {
  return value instanceof Placeholder ? value.index : -1;
}
function verifyPlaceholder(placeholder, value) {
  return placeholder.verifyValue(value);
}
function setObjectTemplatePlaceholder(template, objectPath, objectPathIndex, placeholder) {
  if (template instanceof CustomObject) {
    template.addPlaceholder(placeholder);
    placeholder.belongsToCustomObject = template;
    return;
  }
  var key = objectPath[objectPathIndex];
  if (objectPathIndex < objectPath.length - 1) {
    return setObjectTemplatePlaceholder(template[key], objectPath, objectPathIndex + 1, placeholder);
  }
  template[key] = placeholder;
}
function setValue(object, objectPath, objectPathIndex, value) {
  var key = objectPath[objectPathIndex];
  if (objectPathIndex < objectPath.length - 1) {
    return setValue(object[key], objectPath, objectPathIndex + 1, value);
  }
  object[key] = value;
}
function getValue(object, objectPath, objectPathIndex) {
  var key = objectPath[objectPathIndex];
  if (objectPathIndex < objectPath.length - 1) {
    return getValue(object[key], objectPath, objectPathIndex + 1);
  }
  return object[key];
}
var Entry = function() {
  function Entry(urlTemplate, objectTemplate, options) {
    this.urlTemplate = null;
    this.objectTemplate = null;
    this.urlMatcher = null;
    this.urlComponents = null;
    this.placeholders = null;
    this.customObjects = null;
    this._compile(urlTemplate, objectTemplate, options);
  }
  return ($traceurRuntime.createClass)(Entry, {
    _compile: function(urlTemplate, objectTemplate, options) {
      var normalize = !options || options.normalize !== false;
      urlTemplate = String(urlTemplate);
      objectTemplate = clone(Object(objectTemplate), false);
      var urlMatcher = '';
      var urlMatcherLastGroup = 0;
      var placeholderContentBuffer = null;
      var state = PARSE_STATE.PLAIN;
      var urlComponents = [];
      var placeholders = [];
      var customObjects = [];
      if (normalize) {
        urlTemplate = normalizeUrl(urlTemplate);
      }
      initializeCustomObjects(objectTemplate, customObjects, []);
      for (var i = 0; i < urlTemplate.length; ++i) {
        var c = urlTemplate[i];
        if (state === PARSE_STATE.PLAIN) {
          if (c === '{') {
            state = PARSE_STATE.PLACEHOLDER;
            placeholderContentBuffer = '';
          } else {
            urlMatcher += escapeRegExp(c);
            if (typeof urlComponents[urlComponents.length - 1] === 'string') {
              urlComponents[urlComponents.length - 1] += c;
            } else {
              urlComponents.push(c);
            }
          }
        } else if (state === PARSE_STATE.PLACEHOLDER) {
          if (c === '}') {
            urlMatcher += matcherPlaceholderDefault;
            ++urlMatcherLastGroup;
            var placeholder = new Placeholder();
            placeholder.objectPath = placeholderContentBuffer.split('.');
            placeholder.urlMatcherGroupIndex = urlMatcherLastGroup;
            placeholder.verifyMode = VERIFY.NO_SLASH;
            placeholder.index = placeholders.length;
            setObjectTemplatePlaceholder(objectTemplate, placeholder.objectPath, 0, placeholder);
            placeholders.push(placeholder);
            urlComponents.push(placeholder);
            state = PARSE_STATE.PLAIN;
            placeholderContentBuffer = null;
          } else {
            placeholderContentBuffer += c;
          }
        } else {
          throw Error('Assertion error');
        }
      }
      this.urlMatcher = new RegExp('^' + urlMatcher + '$');
      this.urlTemplate = urlTemplate;
      this.objectTemplate = objectTemplate;
      this.urlComponents = urlComponents;
      this.placeholders = placeholders;
      this.customObjects = customObjects;
    },
    fromURL: function(url, options) {
      var normalize = !options || options.normalize !== false;
      if (normalize) {
        url = normalizeUrl(url);
      }
      var match = this.urlMatcher.exec(url);
      if (!match) {
        return null;
      }
      var object = clone(this.objectTemplate);
      var $__7 = true;
      var $__8 = false;
      var $__9 = undefined;
      try {
        for (var $__5 = void 0,
            $__4 = (this.placeholders)[$traceurRuntime.toProperty(Symbol.iterator)](); !($__7 = ($__5 = $__4.next()).done); $__7 = true) {
          var placeholder = $__5.value;
          {
            if (placeholder.belongsToCustomObject) {
              continue;
            }
            var value = match[placeholder.urlMatcherGroupIndex];
            value = decodeURIComponent(value);
            setValue(object, placeholder.objectPath, 0, value);
          }
        }
      } catch ($__10) {
        $__8 = true;
        $__9 = $__10;
      } finally {
        try {
          if (!$__7 && $__4.return != null) {
            $__4.return();
          }
        } finally {
          if ($__8) {
            throw $__9;
          }
        }
      }
      var $__21 = true;
      var $__22 = false;
      var $__23 = undefined;
      try {
        for (var $__19 = void 0,
            $__18 = (this.customObjects)[$traceurRuntime.toProperty(Symbol.iterator)](); !($__21 = ($__19 = $__18.next()).done); $__21 = true) {
          var customObject = $__19.value;
          {
            var values = {};
            var $__14 = true;
            var $__15 = false;
            var $__16 = undefined;
            try {
              for (var $__12 = void 0,
                  $__11 = (customObject.placeholders)[$traceurRuntime.toProperty(Symbol.iterator)](); !($__14 = ($__12 = $__11.next()).done); $__14 = true) {
                var placeholder$__25 = $__12.value;
                {
                  var value$__26 = match[placeholder$__25.urlMatcherGroupIndex];
                  value$__26 = decodeURIComponent(value$__26);
                  setValue(values, placeholder$__25.objectPath, customObject.objectPath.length, value$__26);
                }
              }
            } catch ($__17) {
              $__15 = true;
              $__16 = $__17;
            } finally {
              try {
                if (!$__14 && $__11.return != null) {
                  $__11.return();
                }
              } finally {
                if ($__15) {
                  throw $__16;
                }
              }
            }
            var customObjectValue = customObject.create(values);
            setValue(object, customObject.objectPath, 0, customObjectValue);
          }
        }
      } catch ($__24) {
        $__22 = true;
        $__23 = $__24;
      } finally {
        try {
          if (!$__21 && $__18.return != null) {
            $__18.return();
          }
        } finally {
          if ($__22) {
            throw $__23;
          }
        }
      }
      return object;
    },
    toURL: function(object) {
      var placeholderValues = new Array(this.placeholders.length);
      var match = compareObject(this.objectTemplate, object, {
        verifyValue: verifyCompareObjectValue,
        values: placeholderValues,
        indexOfPlaceholder: indexOfPlaceholder,
        verifyPlaceholder: verifyPlaceholder
      });
      if (!match) {
        return null;
      }
      var url = '';
      var customObjectValues = new Array(this.customObjects.length);
      for (var i = 0; i < this.customObjects.length; ++i) {
        var customObjectTemplate = this.customObjects[i];
        var customObject = getValue(object, customObjectTemplate.objectPath, 0);
        customObjectValues[i] = customObjectTemplate.serializeValue(customObject);
      }
      for (var i$__27 = 0; i$__27 < this.urlComponents.length; ++i$__27) {
        var component = this.urlComponents[i$__27];
        if (component instanceof Placeholder) {
          var placeholder = component;
          var value;
          if (placeholder.belongsToCustomObject) {
            var customObject$__28 = placeholder.belongsToCustomObject;
            var values = customObjectValues[customObject$__28.index];
            value = getValue(values, placeholder.objectPath, customObject$__28.objectPath.length);
          } else {
            value = placeholderValues[component.index];
          }
          value = encodeURIComponent(value);
          url += value;
        } else {
          url += component;
        }
      }
      return url;
    }
  }, {});
}();
module.exports = Entry;
